import random
from datetime import timedelta
import numpy as np
from pygdg.common import *

def random_gauss_clamp(mu, sigma, factor=3):
    return min(mu+factor*sigma, max(mu-factor*sigma, random.gauss(mu, sigma)))

class Interpolator:
    def __init__(self, dictionary, mod = False):
        self.keys = []
        self.values = []
        self.mod = mod

        for key in dictionary:
            self.keys.append(float(key))
            self.values.append(float(dictionary[key]))

    def __getitem__(self, key):
        mod_key = key
        if mod_key > self.keys[-1] and self.mod:
            mod_key = key % self.keys[-1]
        return np.interp(key, self.keys, self.values)

    def __mul__(self, value):
        return Interpolator(dict(zip(self.keys, (v * value for v in self.values))))

    __rmul__ = __mul__

class SessionOptions:
    """A session options class."""

    def __init__(self, time_mu, time_sigma, 
                        duration_mu, duration_sigma):
        self.time_mu = time_mu
        self.time_sigma = time_sigma
        self.duration_mu = duration_mu
        self.duration_sigma = duration_sigma

    def time(self):
        return timedelta(seconds=random.gauss(self.time_mu.total_seconds(), self.time_sigma.total_seconds()))

    def duration(self):
        return timedelta(seconds=random_gauss_clamp(self.duration_mu.total_seconds(), self.duration_sigma.total_seconds()))

class PlayerOptions:
    """A player options class."""   
     
    def __init__(self, player_type, 
                        sessions_options,
                        lifetime):
        self.player_type = player_type
        self.sessions_options = sessions_options
        self.lifetime = lifetime

class GameOptions:
    """A game options class."""

    def __init__(self, players_options, players_acquisition, simulation_days):
        self.players_options = {k: v for k, v in sorted(players_options.items(), key=lambda item: item[1])}
        self.players_acquisition = players_acquisition
        self.simulation_days = simulation_days

    def __getitem__(self, p):
        for key, value in self.players_options.items():
            if p < value:
                return key
            else:
                last_key = key
        return last_key

def default_game_options(players, days):

    morning_session_options = SessionOptions(
        timedelta(hours=7), # session time mean
        timedelta(minutes=30), # session time standard deviation
        timedelta(minutes=30), # session duration mean
        timedelta(minutes=10), # session duration standard deviation
    )

    noon_session_options = SessionOptions(
        timedelta(hours=12), # session time mean
        timedelta(hours=30), # session time standard deviation
        timedelta(minutes=45), # session duration mean
        timedelta(minutes=15), # session duration standard deviation
    )

    afternoon_session_options = SessionOptions(
        timedelta(hours=17), # session time mean
        timedelta(hours=1), # session time standard deviation
        timedelta(minutes=30), # session duration mean
        timedelta(minutes=20), # session duration standard deviation
    )

    night_session_options = SessionOptions(
        timedelta(hours=20), # session time mean
        timedelta(hours=3), # session time standard deviation
        timedelta(hours=2), # session duration mean
        timedelta(minutes=1), # session duration standard deviation
    )

    bot_player = PlayerOptions(
        'bot',
        {
            WeekDay.MONDAY: {
                night_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            },
            WeekDay.TUESDAY: {
                morning_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            },
            WeekDay.WEDNESDAY:  {
                night_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            },
            WeekDay.THURSDAY:  {
                night_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            },
            WeekDay.FRIDAY:  {
                night_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            },
            WeekDay.SATURDAY:  {
                night_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            },
            WeekDay.SATURDAY: {
                night_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            },
            WeekDay.SUNDAY: {
                night_session_options: 1.0,
                noon_session_options: 1.0,
                afternoon_session_options: 1.0,
                night_session_options: 1.0
            }
        }, 
        Interpolator({
            0: 1.0
        })
    )

    hardcore_player_options = PlayerOptions(
        'hardcore',
        {
            WeekDay.MONDAY: {
                night_session_options: 0.5  # 50% chance to play a night session on monday
            },
            WeekDay.TUESDAY: {
                morning_session_options: 0.3,
                noon_session_options: 0.5,
                night_session_options: 1.0
            },
            WeekDay.WEDNESDAY:  {
                morning_session_options: 0.3,
                noon_session_options: 0.5,
                night_session_options: 1.0
            },
            WeekDay.THURSDAY:  {
                morning_session_options: 0.3,
                noon_session_options: 0.5,
                night_session_options: 1.0
            },
            WeekDay.FRIDAY:  {
                morning_session_options: 0.3,
                noon_session_options: 0.5,
                night_session_options: 1.0
            },
            WeekDay.SATURDAY:  {
                morning_session_options: 0.3,
                noon_session_options: 0.5,
                night_session_options: 1.0
            },
            WeekDay.SATURDAY: {
                morning_session_options: 0.5
            },
            WeekDay.SUNDAY: {
                afternoon_session_options: 0.8
            }
        }, Interpolator({
            0: 0.7, # day 1 modifier for session probability
            6: 1.0,
            13: 0.7,
            27: 0.5
        })
    )

    casual_player_options = PlayerOptions(
        'casual',
        {
            WeekDay.TUESDAY: {
                noon_session_options: 0.5,
                night_session_options: 0.5
            },
            WeekDay.WEDNESDAY:  {
                noon_session_options: 0.5,
                night_session_options: 0.5
            },
            WeekDay.THURSDAY:  {
                noon_session_options: 0.5,
                night_session_options: 0.5
            },
            WeekDay.FRIDAY:  {
                noon_session_options: 0.5,
                night_session_options: 0.5
            }
        }, Interpolator({
            0: 0.7, # day 1 modifier for session probability
            6: 1.0,
            13: 0.3
        })
    )    

    churner_player_options = PlayerOptions(
        'churner',
        {
            WeekDay.TUESDAY: {
                night_session_options: 1.0
            },
            WeekDay.WEDNESDAY:  {
                night_session_options: 1.0
            },
            WeekDay.THURSDAY:  {
                night_session_options: 1.0
            },
            WeekDay.FRIDAY:  {
                night_session_options: 1.0
            },
            WeekDay.SUNDAY:  {
                night_session_options: 1.0
            }
        }, Interpolator({
            0: 1.0, # day 1 modifier for session probability
            1: 1.0,
            2: 0.5,
            6: 0.3,
            13: 0
        })
    )

    game_options = GameOptions(
        players_options = {
            # bot_player: 1.0,
            hardcore_player_options: 0.05,
            casual_player_options: 0.1,
            churner_player_options: 1.0,
        }, 
        players_acquisition = players * Interpolator({
                0: 1, # one player acquired day one
                6: 2,
                7: 0,
    
                13: 0,
                14: 2,
                20: 5,
                21: 0,
    
                27: 0,
                28: 1,
                34: 3,
                35: 0
            }, 
            mod=True
        ),
        simulation_days = days # simulation days
    )

    return game_options


from os.path import exists
import random
import pandas as pd
from enum import Enum
from functools import reduce, partial
from pygdg.common import *

ONE_MINUTE_IN_SECONDS = 60
ONE_HOUR_IN_SECONDS = ONE_MINUTE_IN_SECONDS * 60
ONE_DAY_IN_SECONDS = ONE_HOUR_IN_SECONDS * 24
ONE_WEEK_IN_SECONDS = ONE_DAY_IN_SECONDS * 7
ONE_MONTH_IN_SECONDS = ONE_WEEK_IN_SECONDS * 4

class Counter:
    def __init__(self, count, total):
        self.count = count
        self.total = total

    def __iadd__(self, increment):
        self.count += increment
        return self    

class FeatureName(Enum):
    cohort_id = 0
    cohort_day_of_week = 1
    player_id = 2
    player_type = 3
    player_lifetime = 4
    player_churn = 5
    session_count = 6
    last_minute = 7
    last_hour = 8
    last_day = 9
    last_week = 10
    last_month = 11

    @classmethod
    def last_minute_suffix(cls, minute, prefix):
        return f'_{prefix}_{cls.last_minute.name}(-{minute})'

    @classmethod        
    def last_hour_suffix(cls, hour, prefix):
        return f'_{prefix}_{cls.last_hour.name}(-{hour})'        

    @classmethod
    def last_day_suffix(cls, day, prefix):
        return f'_{prefix}_{cls.last_day.name}(-{day})'

    @classmethod
    def last_week_suffix(cls, week, prefix):
        return f'_{prefix}_{cls.last_week.name}(-{week})'

    @classmethod
    def last_month_suffix(cls, month, prefix):
        return f'_{prefix}_{cls.last_month.name}(-{month})'

    @classmethod
    def names(cls):
        return list(map(lambda e: e.name, cls))

class FeatureVariant(Enum):
    count = 0
    time_of_day_mean = 1
    time_of_day_std = 2

    @classmethod
    def names(cls):
        return list(map(lambda e: e.name, cls))

def extract_cohort_id(player_events):
    first_session_timestamp = player_events[PlayerEventField.timestamp.name].iat[0]
    return {FeatureName.cohort_id.name: first_session_timestamp.strftime("%Y_%m_%d")}

def extract_cohort_day_of_week(player_events):
    first_session_timestamp = player_events[PlayerEventField.timestamp.name].iat[0]
    return {FeatureName.cohort_day_of_week.name: first_session_timestamp.day_of_week}   

def extract_player_type(player_events):
    player_type = player_events[PlayerEventField.player_type.name].iat[0]
    return {FeatureName.player_type.name: player_type}

def extract_player_session_count(player_events):
    player_session_count = player_events[PlayerEventField.session_id.name].drop_duplicates().size
    return {FeatureName.session_count.name: player_session_count}

def extract_player_lifetime(player_events):
    first_session_timestamp = player_events[PlayerEventField.timestamp.name].iat[0]
    last_session_timestamp = player_events[PlayerEventField.timestamp.name].iat[-1]
    player_lifetime = (last_session_timestamp - first_session_timestamp).total_seconds()
    return {FeatureName.player_lifetime.name: player_lifetime}  

def extract_player_churn(player_events, timestamp, days):
    features = dict()
    
    last_session_timestamp = player_events[PlayerEventField.timestamp.name].iat[-1]
    inactive_days = (timestamp - last_session_timestamp).days
    if inactive_days > days:
        features[FeatureName.player_churn.name] = True
    else:
        features[FeatureName.player_churn.name] = False

    return features

def extract_player_events_by_time_periods(player_events, minutes, hours, days, weeks, months, operator, prefix):
    
    def extract_player_events_by_time_period(player_events_by_elapsed_time_periods, time_period, suffixer):
        features = dict()
    
        for time in range(0, time_period):
            feature_suffix = suffixer(time+1)
            if time in player_events_by_elapsed_time_periods.groups:
                player_events_by_elapsed_time_period = player_events_by_elapsed_time_periods.get_group(time).groupby(PlayerEventField.event_type.name)
                for event_type in PlayerEventType:
                    feature_name = f'{event_type.name.lower()}{feature_suffix}'
                    if event_type.name in player_events_by_elapsed_time_period.groups:  
                        player_events_by_elapsed_time_period_and_event_type = player_events_by_elapsed_time_period.get_group(event_type.name)
                        features[feature_name] = operator(player_events_by_elapsed_time_period_and_event_type)
                    else:
                        features[feature_name] = 0
            else:
                for event_type in PlayerEventType:
                    features[f'{event_type.name.lower()}{feature_suffix}'] = 0
        
        return features
    
    features = dict()

    last_session_timestamp = player_events[PlayerEventField.timestamp.name].iat[-1]

    elapsed_time_in_seconds = 'elapsed_time_in_seconds'
    elapsed_time_in_minutes = 'elapsed_time_in_minutes'
    elapsed_time_in_hours = 'elapsed_time_in_hours'
    elapsed_time_in_days = 'elapsed_time_in_days'
    elapsed_time_in_weeks = 'elapsed_time_in_weeks' 
    elapsed_time_in_months = 'elapsed_time_in_months' 

    player_events[elapsed_time_in_seconds] = player_events[PlayerEventField.timestamp.name].apply(lambda d: int((last_session_timestamp-d).total_seconds()))
    player_events[elapsed_time_in_minutes] = player_events[elapsed_time_in_seconds] // ONE_MINUTE_IN_SECONDS 
    player_events[elapsed_time_in_hours] = player_events[elapsed_time_in_seconds] // ONE_HOUR_IN_SECONDS
    player_events[elapsed_time_in_days] = player_events[elapsed_time_in_seconds] // ONE_DAY_IN_SECONDS
    player_events[elapsed_time_in_weeks] = player_events[elapsed_time_in_seconds] // ONE_WEEK_IN_SECONDS
    player_events[elapsed_time_in_months] = player_events[elapsed_time_in_seconds] // ONE_MONTH_IN_SECONDS

    features.update(extract_player_events_by_time_period(player_events.groupby(elapsed_time_in_minutes), minutes, partial(FeatureName.last_minute_suffix, prefix=prefix)))
    features.update(extract_player_events_by_time_period(player_events.groupby(elapsed_time_in_hours), hours, partial(FeatureName.last_hour_suffix, prefix=prefix)))
    features.update(extract_player_events_by_time_period(player_events.groupby(elapsed_time_in_days), days, partial(FeatureName.last_day_suffix, prefix=prefix)))
    features.update(extract_player_events_by_time_period(player_events.groupby(elapsed_time_in_weeks), weeks, partial(FeatureName.last_week_suffix, prefix=prefix)))
    features.update(extract_player_events_by_time_period(player_events.groupby(elapsed_time_in_months), months, partial(FeatureName.last_month_suffix, prefix=prefix)))

    return features

def extract_player_events_count(player_events_by_elapsed_time_period_and_event_type):
    return player_events_by_elapsed_time_period_and_event_type[PlayerEventField.id.name].count()

def extract_player_events_time_of_day_mean(player_events_by_elapsed_time_period_and_event_type):
    timestamp = player_events_by_elapsed_time_period_and_event_type[PlayerEventField.timestamp.name].mean()
    timedelta = timestamp - pd.Timestamp(year=timestamp.year, month=timestamp.month, day=timestamp.day)
    return timedelta.total_seconds()
    
def extract_player_events_time_of_day_std(player_events_by_elapsed_time_period_and_event_type):
    return player_events_by_elapsed_time_period_and_event_type[PlayerEventField.timestamp.name].std().total_seconds()

def extract_features(player_events, extractors, counter):
    def extract(features, extractor):   
        extracted_features = extractor(player_events)
        return {**features, **extracted_features}

    counter += 1
    print_progress_bar(counter.count, counter.total+1, prefix = 'generating features:', suffix = '', length = 50)
    
    return pd.Series(reduce(extract, extractors, dict()))

class FeaturesOptions:

    def __init__(self, churn_days, last_minutes, last_hours, lasy_days,
                                last_weeks, last_months):
        self.churn_days = churn_days
        self.last_minutes = last_minutes
        self.last_hours = last_hours
        self.last_days = lasy_days
        self.last_weeks = last_weeks
        self.last_months = last_months

def generate_player_features(game_events, features_options):

    game_events[PlayerEventField.timestamp.name] = pd.to_datetime(game_events[PlayerEventField.timestamp.name])
    game_events = game_events.sort_values(by=[PlayerEventField.timestamp.name])
    game_events_by_player_id = game_events.groupby(PlayerEventField.player_id.name)

    # add player ids

    player_count = len(game_events_by_player_id.indices)
    player_features = pd.DataFrame()
    player_features[FeatureName.player_id.name] = game_events[PlayerEventField.player_id.name].drop_duplicates()
    player_features.set_index(keys=FeatureName.player_id.name, inplace=True)

    # extract features

    churn_timestamp = game_events[PlayerEventField.timestamp.name].iat[-1]
    churn_days = features_options.churn_days

    extract_player_events = partial(
        extract_player_events_by_time_periods, 
        minutes=features_options.last_minutes, 
        hours=features_options.last_hours, 
        days=features_options.last_days, 
        weeks=features_options.last_weeks, 
        months=features_options.last_months, 
    )
    
    counter = Counter(1, player_count)
    
    features_extractor = partial(
        extract_features, 
        extractors=[
            extract_cohort_id,
            extract_cohort_day_of_week,
            extract_player_type,
            extract_player_lifetime,
            extract_player_session_count,
            partial(extract_player_churn, 
                    timestamp=churn_timestamp, 
                    days=churn_days),
            partial(extract_player_events,
                    operator=extract_player_events_count,
                    prefix=FeatureVariant.count.name),
            partial(extract_player_events,
                    operator=extract_player_events_time_of_day_mean,
                    prefix=FeatureVariant.time_of_day_mean.name),
            partial(extract_player_events,
                    operator=extract_player_events_time_of_day_std,
                    prefix=FeatureVariant.time_of_day_std.name)
        ],
        counter = counter
    )
    
    print_progress_bar(counter.count, counter.total+1, prefix = 'generating features:', suffix = '', length = 50)
    
    extracted_features = game_events_by_player_id.apply(features_extractor)
    player_features = pd.merge(player_features, extracted_features, left_index=True, right_index=True)

    return player_features

def generate(filename, events, churn_days, last_minutes, last_hours, 
                last_days, last_weeks, last_months, seed, overwrite, debug):
    
    # set seed

    random.seed(seed)

    # load game events

    print('loading events...')
    events_file = f'{events}.csv'
    if not exists(events_file):
        print(f'{events_file} does not exist!')
        return
    events_dataframe = pd.read_csv(events_file)
    print('events loaded!')

    # generate machine learning features

    features_file = f'{filename}.csv'
    
    if not exists(features_file) or overwrite:
        
        features_options = FeaturesOptions(
            churn_days,
            last_minutes,
            last_hours,
            last_days,
            last_weeks,
            last_months
        )
        features_dataframe = generate_player_features(events_dataframe, features_options)

        print('storing features...')
        features_dataframe.to_csv(features_file)
        print(f'features stored in {features_file}!')
        
    else:
        
        print(f'{features_file} already exists, use --overwrite to replace the current features!')


from datetime import date
from functools import reduce
import click
from pygdg.common import PlayerEventField, PlayerEventType
import pygdg.events as e
import pygdg.features as f

# events

DEFAULT_EVENTS_FILENAME='events'
DEFAULT_EVENTS_DATE=str(date.today())
DEFAULT_EVENTS_PLAYERS=10
DEFAULT_EVENTS_DAYS=7

# metrics

DEFAULT_METRICS_FILENAME='metrics'

# features

DEFAULT_FEATURES_FILENAME='features'
DEFAULT_FEATURES_CHURN_DAYS=5
DEFAULT_FEATURES_LAST_MINUTES=0
DEFAULT_FEATURES_LAST_HOURS=0
DEFAULT_FEATURES_LAST_DAYS=7
DEFAULT_FEATURES_LAST_WEEKS=3
DEFAULT_FEATURES_LAST_MONTHS=2

# common

DEFAULT_SEED=0
DEFAULT_PLOT=False
DEFAULT_OVERWRITE=False
DEFAULT_DEBUG=False

@click.group()
@click.version_option()
def main():
    pass

@main.command(help=f'''
Generate game events of types ({','.join(PlayerEventType.names())}) with fields ({','.join(PlayerEventField.names())}) in a specified csv filename (default={DEFAULT_EVENTS_FILENAME}).
''')
@click.option('--date', type=click.DateTime(formats=["%Y-%m-%d"]), default=DEFAULT_EVENTS_DATE, help='The players acquisition starting date (default=today)')
@click.option('--players', default=DEFAULT_EVENTS_PLAYERS, help=f'The number of daily acquired players (default={DEFAULT_EVENTS_PLAYERS})')
@click.option('--days', default=DEFAULT_EVENTS_DAYS, help=f'The number of acquisition days (default={DEFAULT_EVENTS_DAYS})')
@click.option('--seed', default=DEFAULT_SEED, help=f'The random seed (default={DEFAULT_SEED})')
@click.option('--plot/--no-plot', default=DEFAULT_PLOT, help=f'The plot flag (default={DEFAULT_PLOT})')
@click.option('--overwrite/--no-overwrite', default=DEFAULT_PLOT, help=f'The overwrite flag (default={DEFAULT_OVERWRITE})')
@click.option('--debug/--no-debug', default=DEFAULT_DEBUG, help=f'The debug flag (default={DEFAULT_DEBUG})')
@click.argument('filename', default=DEFAULT_EVENTS_FILENAME)
def events(filename, date, players, days, seed, plot, overwrite, debug):
    e.generate(filename, date, players, days, seed, plot, overwrite, debug)

@main.command(help=f'''
Generate metrics from game events (not implemented yet)
''')
@click.option('--seed', default=DEFAULT_SEED, help=f'The random seed (default={DEFAULT_SEED})')
@click.option('--debug/--no-debug', default=DEFAULT_DEBUG, help=f'The debug flag (default={DEFAULT_DEBUG})')
@click.option('--events', default=DEFAULT_EVENTS_FILENAME, help=f'The csv filename of the input game events (default={DEFAULT_EVENTS_FILENAME})')
@click.argument('filename', default=DEFAULT_METRICS_FILENAME)
def metrics(filename, events, seed, debug):
    click.echo('not implemented yet!')

@main.command(help=f'''
Generate machine learning features ({','.join(f.FeatureName.names())}) with variants ({','.join(f.FeatureVariant.names())}) for each event type ({','.join(f.PlayerEventType.names())}) in a specified csv filename (default={DEFAULT_FEATURES_FILENAME})
''')
@click.option('--churn-days', default=DEFAULT_FEATURES_CHURN_DAYS, help=f'The number of inactivity days to be flagged as churn (default={DEFAULT_FEATURES_CHURN_DAYS})')
@click.option('--last-minutes', default=DEFAULT_FEATURES_LAST_MINUTES, help=f'The number of minutes to sample before last event date (default={DEFAULT_FEATURES_LAST_MINUTES})')
@click.option('--last-hours', default=DEFAULT_FEATURES_LAST_HOURS, help=f'The number of hours to sample before last event date (default={DEFAULT_FEATURES_LAST_HOURS})')
@click.option('--last-days', default=DEFAULT_FEATURES_LAST_DAYS, help=f'The number of days to sample before last event date (default={DEFAULT_FEATURES_LAST_DAYS})')
@click.option('--last-weeks', default=DEFAULT_FEATURES_LAST_WEEKS, help=f'The number of minutes to sample before last event date (default={DEFAULT_FEATURES_LAST_WEEKS})')
@click.option('--last-months', default=DEFAULT_FEATURES_LAST_MONTHS, help=f'The number of months to sample before last event date (default={DEFAULT_FEATURES_LAST_MONTHS})')
@click.option('--events', default=DEFAULT_EVENTS_FILENAME, help=f'The csv filename of the input game events (default={DEFAULT_EVENTS_FILENAME})')
@click.option('--seed', default=DEFAULT_SEED, help=f'The random seed (default={DEFAULT_SEED})')
@click.option('--overwrite/--no-overwrite', default=DEFAULT_PLOT, help=f'The overwrite flag (default={DEFAULT_OVERWRITE})')
@click.option('--debug/--no-debug', default=DEFAULT_DEBUG, help=f'The debug flag (default={DEFAULT_DEBUG})')
@click.argument('filename', default=DEFAULT_FEATURES_FILENAME)
def features(filename, events, churn_days, last_minutes, last_hours, 
                last_days, last_weeks, last_months, 
                seed, overwrite, debug):
    f.generate(filename, events, churn_days, last_minutes, last_hours, 
                last_days, last_weeks, last_months, 
                seed, overwrite, debug)

if __name__ == '__main__':
    main()


from enum import Enum

class WeekDay(Enum):
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6
    
class DayHour(Enum):
    TWELVE_AM = 0
    ONE_AM = 1
    TWO_AM = 2
    THREE_AM = 3
    FOUR_AM = 4
    FIVE_AM = 5
    SIX_AM = 6
    SEVEN_AM = 7
    EIGHT_AM = 8
    NINE_AM = 9
    TEN_AM = 10
    ELEVEN_AM = 11
    TWELVE_PM = 12
    ONE_PM = 13
    TWO_PM = 14
    THREE_PM = 15
    FOUR_PM = 16
    FIVE_PM = 17
    SIX_PM = 18
    SEVEN_PM = 19
    EIGHT_PM = 20
    NINE_PM = 21
    TEN_PM = 22
    ELEVEN_PM = 23

class PlayerEventType(Enum):
    BEGIN_SESSION = 0
    END_SESSION = 1
    #BEGIN_STAGE = 2
    #END_STAGE = 3
    #SEND_MESSAGE = 4
    #RECEIVE_MESSAGE = 5
    #ASK_FRIEND = 6
    #ACCEPT_FRIEND = 7
    #VIEW_LEADERBOARD = 8
    #BUY_CURRENCY = 9
    #BUY_ITEM = 10

    @classmethod
    def names(cls):
        return list(map(lambda e: e.name, cls))

class PlayerEventField(Enum):
    id = 0
    player_id = 1
    player_type = 2
    cohort_id = 3
    session_id = 4
    event_type = 5
    timestamp = 6
    #payload = 7
    #stage_id = 6
    #stage_result = 7
    #chat_message = 8
    #friend_id = 9
    #leaderboard_id = 10
    #currency_amount = 11
    #item_id = 12

    @classmethod
    def names(cls):
        return list(map(lambda e: e.name, cls))

def print_progress_bar(iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end = printEnd)
    if iteration == total: 
        print()




from os.path import exists
import random, uuid
from datetime import datetime, timedelta
import pandas as pd
from pygdg.common import *
from pygdg.options import *

class PlayerEvent:
    """A player event class. It represents an event generated by a player during a game session."""
    
    def __init__(self, cohort_id, player_id, player_type, session_id, event_type, timestamp):
        self.id = uuid.uuid4()
        self.cohort_id = cohort_id
        self.player_id = player_id
        self.player_type = player_type
        self.session_id = session_id
        self.event_type = event_type
        self.timestamp = timestamp
    
    def to_dict(self):
        return {
            PlayerEventField.id.name : [self.id.hex],
            PlayerEventField.cohort_id.name : [self.cohort_id.hex],
            PlayerEventField.player_id.name : [self.player_id.hex],
            PlayerEventField.player_type.name : [self.player_type],
            PlayerEventField.session_id.name : [self.session_id.hex],
            PlayerEventField.event_type.name : [self.event_type.name],
            PlayerEventField.timestamp.name : [self.timestamp.isoformat()],
        }

    def to_dataframe(self):
        return pd.DataFrame.from_dict(self.to_dict())

class SessionActivity:
 
    def __init__(self, cohort_id, player_id, player_type, session_id, session_start_date, session_options):
        self.cohort_id = cohort_id
        self.player_id = player_id
        self.player_type = player_type
        self.session_id = session_id
        self.session_start_date = session_start_date
        self.session_options = session_options

    def generate_events(self):

        session_datetime = self.session_start_date + self.session_options.time()
        session_duration = self.session_options.duration()

        begin_session = PlayerEvent(
            self.cohort_id, 
            self.player_id, 
            self.player_type, 
            self.session_id, 
            PlayerEventType.BEGIN_SESSION, 
            session_datetime
        )

        # TODO: add events

        end_session = PlayerEvent(
            self.cohort_id, 
            self.player_id, 
            self.player_type, 
            self.session_id, 
            PlayerEventType.END_SESSION, 
            session_datetime + session_duration
        )
        
        events = [
            begin_session, 
            end_session
        ]

        return pd.concat(map(PlayerEvent.to_dataframe, events))        

class PlayerActivity:

    def __init__(self, cohort_id, player_id, player_type, player_options, player_start_date):
        self.cohort_id = cohort_id
        self.player_id = player_id
        self.player_type = player_type
        self.player_options = player_options
        self.player_start_date = player_start_date
        self.current_day = 0

    def generate_events(self):

        lifetime_weight = self.player_options.lifetime[self.current_day]

        if (lifetime_weight <= 0):
            return None

        sessions_dataframes = []
        session_date = datetime.combine(
            self.player_start_date.date() + timedelta(days=self.current_day), 
            datetime.min.time()
        )
        session_weekday = WeekDay(session_date.weekday())

        if session_weekday in self.player_options.sessions_options:

            sessions_options = self.player_options.sessions_options[session_weekday]

            for session_options in sessions_options:

                weight = lifetime_weight * sessions_options[session_options]

                if (random.random() < weight):
                    session_id = uuid.uuid4()
                    session_activity = SessionActivity(
                        self.cohort_id, 
                        self.player_id, 
                        self.player_type, 
                        session_id, 
                        session_date, 
                        session_options
                    )
                    session_dataframe = session_activity.generate_events()
                    sessions_dataframes.append(session_dataframe)

        self.current_day += 1

        if len(sessions_dataframes) == 0:
            return pd.DataFrame()
        else:
            return pd.concat(sessions_dataframes)  

class GameActivity:

    def __init__(self, game_options, start_date):
        self.game_options = game_options
        self.start_date = start_date
        self.current_day = 0

    def generate_events(self):

        # update progress bar
        
        print_progress_bar(self.current_day, self.game_options.simulation_days, prefix = 'generating events:', suffix = '', length = 50)

        player_activities = []
        players_dataframes = []

        while self.current_day < self.game_options.simulation_days:

            # handle old players activities

            old_players_activities = player_activities
            player_activities = []

            for player_activity in old_players_activities:

                player_dataframe = player_activity.generate_events()

                if player_dataframe is not None:
                    players_dataframes.append(player_dataframe)
                    player_activities.append(player_activity)

            # handle new players activities

            current_new_player = int(self.game_options.players_acquisition[self.current_day])

            cohort_id =uuid.uuid4()

            while current_new_player > 0:

                player_options = self.game_options[random.random()]

                player_id = uuid.uuid4()
                player_type = player_options.player_type
                player_start_date = self.start_date + timedelta(days=self.current_day)

                player_activity = PlayerActivity(
                    cohort_id,
                    player_id, 
                    player_type,
                    player_options, 
                    player_start_date
                )
                player_dataframe = player_activity.generate_events()

                if player_dataframe is not None:
                    players_dataframes.append(player_dataframe)
                    player_activities.append(player_activity)

                current_new_player -= 1
        
            self.current_day += 1

            # update progress bar

            print_progress_bar(self.current_day, self.game_options.simulation_days, prefix = 'generating events:', suffix = '', length = 50)

        return pd.concat(players_dataframes)  

def generate(filename, date, players, days, seed, plot, overwrite, debug):

    # set seed

    random.seed(seed)

    # generate events
    
    events_file = f'{filename}.csv'
    
    if not exists(events_file) or overwrite:
        
        game_activity = GameActivity(
            default_game_options(players, days), 
            date
        )
    
        events_dataframe = game_activity.generate_events()
    
        if events_dataframe.size > 0:
            events_dataframe[PlayerEventField.timestamp.name] = pd.to_datetime(events_dataframe[PlayerEventField.timestamp.name])
            events_dataframe = events_dataframe.sort_values(by=[PlayerEventField.timestamp.name])
            
        print('storing events...')
        events_dataframe.to_csv(events_file, index=False)
        print(f'events stored in {events_file}!')
        
    else:
    
        print(f'{events_file} already exists, use --overwrite to replace the current events!')

    # plot events

    if plot:
        
        player_index = 'player_index'
        events_by_players = events_dataframe.groupby([PlayerEventField.player_id.name])
        
        for index, (player_id, group) in enumerate(events_by_players):
            events_dataframe.loc[events_dataframe[PlayerEventField.player_id.name] == player_id, player_index] = int(index)

        first_event_timestamp = events_dataframe[PlayerEventField.timestamp.name].min()
        last_event_timestamp = events_dataframe[PlayerEventField.timestamp.name].max()
        first_event_datetime = datetime(first_event_timestamp.year, first_event_timestamp.month, first_event_timestamp.day)
        last_event_datetime = datetime(last_event_timestamp.year, last_event_timestamp.month, last_event_timestamp.day+1)
        first_player_index = int(events_dataframe[player_index].min())
        last_player_index = int(events_dataframe[player_index].max()) + 2
        
        plot = events_dataframe.plot.scatter(
            x=PlayerEventField.timestamp.name, y=player_index, 
            title='Players events', xlabel="Days", ylabel='Players',
            xticks=pd.date_range(first_event_datetime, last_event_datetime),
            yticks=range(first_player_index, last_player_index, 1 + (last_player_index-first_player_index) // 20),
            rot=20
        )
        
        print('plot events...')
        plot_file = f'{filename}.png'
        plot.get_figure().savefig(plot_file)
        print(f'events plotted in {plot_file}!')


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pygdg",
    version="0.1.6",
    author="GroWaK",
    author_email="me@growak.org",
    description="A simple comand line tool to create game events data for analytics and machine learning use cases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='game machine learning analytics dataset generator',
    url="https://github.com/growak/pygdg",
    project_urls={
        "Project Tracker": "https://github.com/growak/pygdg/projects",
        "Issue Tracker": "https://github.com/growak/pygdg/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'click',
        'numpy',
        'pandas',
        'matplotlib'
    ],
    entry_points='''
        [console_scripts]
        pygdg=pygdg.main:main
    ''',
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)